@@ -11,17 +11,18 @@
 //!             if let Command::OTHER(line) = command {
 //!                 print!("{}", line);
 //!             }
-//!             if let Command::PRIVMSG(channel, message) = command {
-//!                println!("PRIVMSG received: {} {}", channel, message);
+//!             if let Command::PRIVMSG(nick, channel, message) = command {
+//!                println!("PRIVMSG received from {}: {} {}", nick, channel, message);
 //!             }
 //!         }
 //!         # break;
 //!     }
-//!
 //!     # Ok(())
 //! }
 
 #![warn(missing_docs)]
+use native_tls::TlsConnector;
+
 use std::borrow::Cow;
 use std::fs::File;
 use std::io::{Error, Read, Write};
@@ -30,10 +31,14 @@ use std::path::Path;
 
 use serde_derive::Deserialize;
 
+/// IRC comamnds
+pub mod commands;
+
 /// An IRC client
 pub struct Client {
     config: Config,
-    stream: TcpStream,
+    stream: Option<TcpStream>,
+    sslstream: Option<native_tls::TlsStream<TcpStream>>,
 }
 
 /// Config for the IRC client
@@ -44,97 +49,12 @@ pub struct Config {
     mode: Option<String>,
     nickname: Option<String>,
     port: u16,
+    ssl: bool,
     username: String,
 }
 
-#[doc(hidden)]
-#[derive(Debug)]
-pub enum CapMode {
-    LS,
-    END,
-}
-
-/// IRC commands
-/// Not reccomended to use, use the helper functions instead
-#[derive(Debug)]
-pub enum Command {
-    #[doc(hidden)]
-    CAP(CapMode),
-    /// Joins a channel
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::JOIN("#main".to_string()))?;
-    /// # Ok::<(), std::io::Error>(())
-    /// ```
-    JOIN(
-        /// Channel
-        String,
-    ),
-    /// Sets the mode of the user
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::MODE("#main".to_string(), Some("+B".to_string())))?;
-    /// # Ok::<(), std::io::Error>(())
-    /// ```
-    /// If the MODE is not given (e.g. None), then the client will send "MODE target"
-    MODE(
-        /// Channel
-        String,
-        /// Mode
-        Option<String>,
-    ),
-    #[doc(hidden)]
-    NICK(String),
-    /// Everything that is not a command
-    OTHER(String),
-    #[doc(hidden)]
-    PING(String),
-    #[doc(hidden)]
-    PONG(String),
-    /// Sends a message in a channel
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::PRIVMSG("#main".to_string(), "This is an example message".to_string()))?;
-    /// # Ok::<(), std::io::Error>(())
-    /// ```
-    PRIVMSG(
-        /// Channel
-        String,
-        /// Message
-        String,
-    ),
-    #[doc(hidden)]
-    USER(String, String, String, String),
-}
-
-impl Command {
-    fn from_str(s: &str) -> Self {
-        let new = s.trim();
-
-        if new.starts_with("PING") {
-            let command = new.split_whitespace().collect::<Vec<&str>>()[1].to_string();
-            return Self::PONG(command);
-        } else if new.contains("PRIVMSG") {
-            let parts: Vec<&str> = new.split_whitespace().collect();
-
-            let target = parts[2];
-            let mut builder = String::new();
-            for part in parts[3..].to_vec() {
-                builder.push_str(&format!("{} ", part));
-            }
-
-            return Self::PRIVMSG(target.to_string(), (&builder[1..]).to_string());
-        }
-
-        Self::OTHER(new.to_string())
-    }
-}
-
 impl Client {
-    /// Creates a new client with a given config
+    /// Creates a new client with a given [`Config`].
     /// ```no_run
     /// # use circe::*;
     /// # let config = Config::from_toml("config.toml")?;
@@ -142,24 +62,42 @@ impl Client {
     /// # Ok::<(), std::io::Error>(())
     /// ```
     ///
-    /// Errors if the client could not connect to the given host.
+    /// Returns error if the client could not connect to the host.
     pub fn new(config: Config) -> Result<Self, Error> {
         let stream = TcpStream::connect(format!("{}:{}", config.host, config.port))?;
-        Ok(Self { stream, config })
+        let sslstream: native_tls::TlsStream<TcpStream>;
+
+        if config.ssl {
+            let connector = TlsConnector::new().unwrap();
+            sslstream = connector.connect(config.host.as_str(), stream).unwrap();
+
+            return Ok(Self {
+                config,
+                stream: None,
+                sslstream: Some(sslstream),
+            });
+        } else {
+            return Ok(Self {
+                config,
+                stream: Some(stream),
+                sslstream: None,
+            });
+        }
     }
 
-    /// Identify user and join the specified channels
+    /// Identify user and joins the in the [`Config`] specified channels.
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
     /// client.identify()?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
-    /// Errors if the client could not write to the stream.
+    /// Returns error if the client could not write to the stream.
     pub fn identify(&mut self) -> Result<(), Error> {
-        self.write_command(Command::CAP(CapMode::END))?;
-        self.write_command(Command::USER(
+        self.write_command(commands::Command::CAP(commands::CapMode::LS))?;
+        self.write_command(commands::Command::CAP(commands::CapMode::END))?;
+
+        self.write_command(commands::Command::USER(
             self.config.username.clone(),
             "*".into(),
             "*".into(),
@@ -167,16 +105,18 @@ impl Client {
         ))?;
 
         if let Some(nick) = self.config.nickname.clone() {
-            self.write_command(Command::NICK(nick.to_string()))?;
+            self.write_command(commands::Command::NICK(nick.to_string()))?;
         } else {
-            self.write_command(Command::NICK(self.config.username.clone()))?;
+            self.write_command(commands::Command::NICK(self.config.username.clone()))?;
         }
 
         loop {
             if let Ok(ref command) = self.read() {
                 match command {
-                    Command::PING(code) => self.write_command(Command::PONG(code.to_string()))?,
-                    Command::OTHER(line) => {
+                    commands::Command::PING(code) => {
+                        self.write_command(commands::Command::PONG(code.to_string()))?;
+                    }
+                    commands::Command::OTHER(line) => {
                         if line.contains("001") {
                             break;
                         }
@@ -187,9 +127,9 @@ impl Client {
         }
 
         let config = self.config.clone();
-        self.write_command(Command::MODE(config.username, config.mode))?;
+        self.write_command(commands::Command::MODE(config.username, config.mode))?;
         for channel in config.channels.iter() {
-            self.write_command(Command::JOIN(channel.to_string()))?;
+            self.write_command(commands::Command::JOIN(channel.to_string()))?;
         }
 
         Ok(())
@@ -198,15 +138,22 @@ impl Client {
     fn read_string(&mut self) -> Option<String> {
         let mut buffer = [0u8; 512];
 
-        match self.stream.read(&mut buffer) {
-            Ok(_) => {}
-            Err(_) => return None,
-        };
+        if self.config.ssl {
+            match self.sslstream.as_mut().unwrap().read(&mut buffer) {
+                Ok(_) => {}
+                Err(_) => return None,
+            };
+        } else {
+            match self.stream.as_mut().unwrap().read(&mut buffer) {
+                Ok(_) => {}
+                Err(_) => return None,
+            };
+        }
 
         Some(String::from_utf8_lossy(&buffer).into())
     }
 
-    /// Read data coming from the IRC as a [`Command`]
+    /// Read data coming from the IRC as a [`commands::Command`].
     /// ```no_run
     /// # use circe::*;
     /// # fn main() -> Result<(), std::io::Error> {
@@ -220,11 +167,19 @@ impl Client {
     /// # Ok::<(), std::io::Error>(())
     /// # }
     /// ```
-    ///
-    /// Errors if there are no new messages. This should not be taken as an actual Error, because nothing went wrong.
-    pub fn read(&mut self) -> Result<Command, ()> {
+    /// Returns error if there are no new messages. This should not be taken as an actual error, because nothing went wrong.
+    pub fn read(&mut self) -> Result<commands::Command, ()> {
         if let Some(string) = self.read_string() {
-            return Ok(Command::from_str(&string));
+            let command = commands::Command::from_str(&string);
+
+            if let commands::Command::PONG(command) = command {
+                if let Err(_e) = self.write_command(commands::Command::PONG(command)) {
+                    return Err(());
+                }
+                return Ok(commands::Command::PONG("".to_string()));
+            }
+
+            return Ok(command);
         }
 
         Err(())
@@ -233,37 +188,77 @@ impl Client {
     fn write(&mut self, data: &str) -> Result<(), Error> {
         let formatted = {
             let new = format!("{}\r\n", data);
-
             Cow::Owned(new) as Cow<str>
         };
-        self.stream.write(formatted.as_bytes())?;
+
+        if self.config.ssl {
+            self.sslstream
+                .as_mut()
+                .unwrap()
+                .write(formatted.as_bytes())
+                .unwrap();
+        } else {
+            self.stream.as_mut().unwrap().write(formatted.as_bytes())?;
+        }
 
         Ok(())
     }
 
-    /// Send a [`Command`] to the IRC
+    /// Send a [`commands::Command`] to the IRC.<br>
+    /// Not reccomended to use, use the helper functions instead.
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::PRIVMSG("#main".to_string(), "Hello".to_string()))?;
+    /// client.write_command(Command::PRIVMSG("".to_string() "#main".to_string(), "Hello".to_string()))?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
-    /// Errors if the stream could not write.
-    pub fn write_command(&mut self, command: Command) -> Result<(), Error> {
-        use Command::*;
+    /// Returns error if the client could not write to the stream.
+    pub fn write_command(&mut self, command: commands::Command) -> Result<(), Error> {
+        use commands::Command::*;
         let computed = match command {
+            ADMIN(target) => {
+                let formatted = format!("ADMIN {}", target);
+                Cow::Owned(formatted) as Cow<str>
+            }
+            AWAY(message) => {
+                let formatted = format!("AWAY {}", message);
+                Cow::Owned(formatted) as Cow<str>
+            }
             CAP(mode) => {
-                use CapMode::*;
+                use commands::CapMode::*;
                 Cow::Borrowed(match mode {
                     LS => "CAP LS 302",
                     END => "CAP END",
                 }) as Cow<str>
             }
+            INVITE(username, channel) => {
+                let formatted = format!("INVITE {} {}", username, channel);
+                Cow::Owned(formatted) as Cow<str>
+            }
             JOIN(channel) => {
                 let formatted = format!("JOIN {}", channel);
                 Cow::Owned(formatted) as Cow<str>
             }
+            LIST(channel, server) => {
+                let mut formatted = "LIST".to_string();
+                if let Some(channel) = channel {
+                    formatted.push_str(format!(" {}", channel).as_str());
+                }
+                if let Some(server) = server {
+                    formatted.push_str(format!(" {}", server).as_str());
+                }
+                Cow::Owned(formatted) as Cow<str>
+            }
+            NAMES(channel, server) => {
+                let formatted = {
+                    if let Some(server) = server {
+                        format!("NAMES {} {}", channel, server)
+                    } else {
+                        format!("NAMES {}", channel)
+                    }
+                };
+                Cow::Owned(formatted) as Cow<str>
+            }
             NICK(nickname) => {
                 let formatted = format!("NICK {}", nickname);
                 Cow::Owned(formatted) as Cow<str>
@@ -276,7 +271,10 @@ impl Client {
                         format!("MODE {}", target)
                     }
                 };
-
+                Cow::Owned(formatted) as Cow<str>
+            }
+            OPER(nick, password) => {
+                let formatted = format!("OPER {} {}", nick, password);
                 Cow::Owned(formatted) as Cow<str>
             }
             OTHER(_) => {
@@ -285,6 +283,14 @@ impl Client {
                     "Cannot write commands of type OTHER",
                 ));
             }
+            PART(target) => {
+                let formatted = format!("PART {}", target);
+                Cow::Owned(formatted) as Cow<str>
+            }
+            PASS(password) => {
+                let formatted = format!("PASS {}", password);
+                Cow::Owned(formatted) as Cow<str>
+            }
             PING(target) => {
                 let formatted = format!("PING {}", target);
                 Cow::Owned(formatted) as Cow<str>
@@ -293,10 +299,24 @@ impl Client {
                 let formatted = format!("PONG {}", code);
                 Cow::Owned(formatted) as Cow<str>
             }
-            PRIVMSG(target, message) => {
+            PRIVMSG(_, target, message) => {
                 let formatted = format!("PRIVMSG {} {}", target, message);
                 Cow::Owned(formatted) as Cow<str>
             }
+            QUIT(message) => {
+                let formatted = format!("QUIT :{}", message);
+                Cow::Owned(formatted) as Cow<str>
+            }
+            TOPIC(channel, topic) => {
+                let formatted = {
+                    if let Some(topic) = topic {
+                        format!("TOPIC {} :{}", channel, topic)
+                    } else {
+                        format!("TOPIC {}", channel)
+                    }
+                };
+                Cow::Owned(formatted) as Cow<str>
+            }
             USER(username, s1, s2, realname) => {
                 let formatted = format!("USER {} {} {} :{}", username, s1, s2, realname);
                 Cow::Owned(formatted) as Cow<str>
@@ -307,77 +327,197 @@ impl Client {
         Ok(())
     }
 
-    // Helper functions
+    // Helper commands
+
+    /// Helper function for requesting information about the ADMIN of an IRC server
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.admin("192.168.178.100")?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn admin(&mut self, target: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::ADMIN(target.to_string()))?;
+        Ok(())
+    }
+
+    /// Helper function for setting the users status to AWAY
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.away("AFK")?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn away(&mut self, message: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::AWAY(message.to_string()))?;
+        Ok(())
+    }
 
     /// Helper function for sending PRIVMSGs.
-    /// This makes
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.send_privmsg("#main", "Hello")?;
+    /// client.privmsg("#main", "Hello")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
-    /// equivalent to
-    ///
+    pub fn privmsg(&mut self, channel: &str, message: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::PRIVMSG(
+            String::from(""),
+            channel.to_string(),
+            message.to_string(),
+        ))?;
+        Ok(())
+    }
+
+    /// Helper function to INVITE people to a channels
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::PRIVMSG("#main".to_string(), "Hello".to_string()))?;
+    /// client.invite("liblirc", "#circe")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    pub fn send_privmsg(&mut self, channel: &str, message: &str) -> Result<(), Error> {
-        self.write_command(Command::PRIVMSG(channel.to_string(), message.to_string()))?;
+    pub fn invite(&mut self, username: &str, channel: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::INVITE(
+            username.to_string(),
+            channel.to_string(),
+        ))?;
         Ok(())
     }
 
     /// Helper function for sending JOINs.
-    /// This makes
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.send_join("#main")?;
+    /// client.join("#main")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
-    /// equivalent to
-    ///
+    pub fn join(&mut self, channel: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::JOIN(channel.to_string()))?;
+        Ok(())
+    }
+
+    /// Helper function for LISTing channels and modes
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::JOIN("#main".to_string()))?;
+    /// client.list(None, None)?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    pub fn send_join(&mut self, channel: &str) -> Result<(), Error> {
-        self.write_command(Command::JOIN(channel.to_string()))?;
+    pub fn list(&mut self, channel: Option<&str>, server: Option<&str>) -> Result<(), Error> {
+        let channel_config = {
+            if let Some(channel) = channel {
+                Some(channel.to_string())
+            } else {
+                None
+            }
+        };
+        let server_config = {
+            if let Some(server) = server {
+                Some(server.to_string())
+            } else {
+                None
+            }
+        };
+        self.write_command(commands::Command::LIST(channel_config, server_config))?;
+        Ok(())
+    }
+
+    /// Helper function for getting all nicknames in a channel
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.names("#main,#circe", None)?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn names(&mut self, channel: &str, server: Option<&str>) -> Result<(), Error> {
+        if let Some(server) = server {
+            self.write_command(commands::Command::NAMES(
+                channel.to_string(),
+                Some(server.to_string()),
+            ))?;
+        } else {
+            self.write_command(commands::Command::NAMES(channel.to_string(), None))?;
+        }
         Ok(())
     }
 
     /// Helper function for sending MODEs.
-    /// This makes
     /// ```no_run
     /// # use circe::*;
     /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.send_mode("test", Some("+B"))?;
+    /// client.mode("test", Some("+B"))?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
-    /// equivalent to
-    ///
-    /// ```no_run
-    /// # use circe::*;
-    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
-    /// client.write_command(Command::MODE("test".to_string(), Some("+B".to_string())))?;
-    /// # Ok::<(), std::io::Error>(())
-    /// ```
-    pub fn send_mode(&mut self, target: &str, mode: Option<&str>) -> Result<(), Error> {
+    pub fn mode(&mut self, target: &str, mode: Option<&str>) -> Result<(), Error> {
         if let Some(mode) = mode {
-            self.write_command(Command::MODE(target.to_string(), Some(mode.to_string())))?;
+            self.write_command(commands::Command::MODE(
+                target.to_string(),
+                Some(mode.to_string()),
+            ))?;
         } else {
-            self.write_command(Command::MODE(target.to_string(), None))?;
+            self.write_command(commands::Command::MODE(target.to_string(), None))?;
         }
         Ok(())
     }
+
+    /// Helper function for leaving channels.
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.part("#main")?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn part(&mut self, target: &str) -> Result<(), Error> {
+        self.write_command(commands::Command::PART(target.to_string()))?;
+        Ok(())
+    }
+
+    /// Helper function for setting or getting the topic of a channel
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.topic("#main", Some("main channel"))?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn topic(&mut self, channel: &str, topic: Option<&str>) -> Result<(), Error> {
+        if let Some(topic) = topic {
+            self.write_command(commands::Command::TOPIC(
+                channel.to_string(),
+                Some(topic.to_string()),
+            ))?;
+        } else {
+            self.write_command(commands::Command::TOPIC(channel.to_string(), None))?;
+        }
+        Ok(())
+    }
+
+    /// Helper function for leaving the IRC server and shutting down the TCP stream afterwards.
+    /// ```no_run
+    /// # use circe::*;
+    /// # let mut client = Client::new(Config::from_toml("config.toml")?)?;
+    /// client.quit(None)?;
+    /// # Ok::<(), std::io::Error>(())
+    /// ```
+    pub fn quit(&mut self, message: Option<&str>) -> Result<(), Error> {
+        if let Some(message) = message {
+            self.write_command(commands::Command::QUIT(message.to_string()))?;
+        } else {
+            self.write_command(commands::Command::QUIT(format!(
+                "circe {} (https://crates.io/crates/circe)",
+                env!("CARGO_PKG_VERSION")
+            )))?;
+        }
+        if self.config.ssl {
+            self.sslstream.as_mut().unwrap().shutdown().unwrap();
+        } else {
+            self.stream
+                .as_mut()
+                .unwrap()
+                .shutdown(std::net::Shutdown::Both)?;
+        }
+
+        Ok(())
+    }
 }
 
 impl Config {
@@ -386,10 +526,10 @@ impl Config {
     /// ```rust
     /// # use circe::*;
     /// let config = Config::new(
-    ///     vec!["#main", "#main2"],
+    ///     vec!["#main", "#circe"],
     ///     "192.168.178.100",
     ///     Some("+B"),
-    ///     Some("IRSC"),
+    ///     Some("circe"),
     ///     6667,
     ///     "circe",
     /// );
@@ -400,21 +540,22 @@ impl Config {
         mode: Option<&'static str>,
         nickname: Option<&'static str>,
         port: u16,
+        ssl: bool,
         username: &str,
     ) -> Self {
         // Conversion from &'static str to String
         let channels_config = channels.iter().map(|channel| channel.to_string()).collect();
 
         let mode_config: Option<String>;
-        if mode.is_some() {
-            mode_config = Some(mode.unwrap().to_string());
+        if let Some(mode) = mode {
+            mode_config = Some(mode.to_string());
         } else {
             mode_config = None;
         }
 
         let nickname_config: Option<String>;
-        if nickname.is_some() {
-            nickname_config = Some(nickname.unwrap().to_string());
+        if let Some(nickname) = nickname {
+            nickname_config = Some(nickname.to_string());
         } else {
             nickname_config = Some(username.to_string());
         }
@@ -425,6 +566,7 @@ impl Config {
             mode: mode_config,
             nickname: nickname_config,
             port,
+            ssl,
             username: username.into(),
         }
     }
@@ -435,7 +577,6 @@ impl Config {
     /// let config = Config::from_toml("config.toml")?;
     /// # Ok::<(), std::io::Error>(())
     /// ```
-    ///
     /// ```toml
     /// channels = ["#main", "#main2"]
     /// host = "192.168.178.100"
@@ -444,7 +585,6 @@ impl Config {
     /// port = 6667
     /// username = "circe"
     /// ```
-    ///
     /// Returns an Error if the file cannot be opened or if the TOML is invalid
     pub fn from_toml<P: AsRef<Path>>(path: P) -> Result<Self, Error> {
         let mut file = File::open(&path)?;
